﻿using System.Collections.Generic;

namespace JPing
{
    /// <summary>
    /// A FIFO collection with amount limit. 
    /// If you enqueue an item when the queue is full, the first item in the queue will be disposed automatically.
    /// </summary>
    /// <typeparam name="T">T</typeparam>
    public class FixedSizeQueue<T> : Queue<T>
    {
        /// <summary>
        /// Lock.
        /// </summary>
        private readonly object _lock = new object();

        /// <summary>
        /// Capacity of queue.
        /// </summary>
        private int _capacity = 10;

        /// <summary>
        /// Get or set the capacity of queue, minimum is 1.
        /// </summary>
        public int Capacity
        {
            get { return _capacity; }
            set
            {
                if (value < 1)
                    _capacity = 1;
                else
                    _capacity = value;

                lock (_lock)
                {
                    while (Count > _capacity)
                        Dequeue();
                }
            }
        }

        /// <summary>
        /// Get the queue is full or not.
        /// </summary>
        public bool IsFull
        {
            get { return Count >= _capacity; }
        }

        /// <summary>
        /// Get the queue is empty or not.
        /// </summary>
        public bool IsEmpty
        {
            get { return Count <= 0; }
        }

        /// <summary>
        /// Create a queue and assign the maximum capacity.
        /// </summary>
        /// <param name="capacity">Maximum capacity</param>
        public FixedSizeQueue(int capacity = 10)
            : base(capacity)
        {
            Capacity = capacity;
        }

        /// <summary>
        /// Add an object in the tail of queue. 
        /// If the queue size reached to the capacity, the oldest object will be dispose automatically.
        /// </summary>
        /// <param name="item">Object</param>
        public new void Enqueue(T item)
        {
            base.Enqueue(item);
            lock (_lock)
            {
                while (Count > _capacity)
                    Dequeue();
            }
        }
    }
}
