﻿using Microsoft.Win32;
using System;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Net.NetworkInformation;
using System.Text;
using System.Timers;
using System.Windows;
using System.Windows.Input;
using System.Windows.Interop;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Threading;
using System.Xml.Serialization;

namespace JPing
{
    /// <summary>
    /// JPing Main UI.
    /// </summary>
    public partial class MainWindow : Window
    {
        #region Constants

        /// <summary>
        /// Path of registry.
        /// </summary>
        private const string REG_PATH = @"HKEY_CURRENT_USER\Software\J.C Studio\JPing";

        /// <summary>
        /// Key of registry.
        /// </summary>
        private const string REG_KEY = @"XSConfig";

        /// <summary>
        /// Ratio between millisecond(ms) and second(s).
        /// </summary>
        private const int MS_2_S = 1000;

        #endregion


        #region Global Variables

        /// <summary>
        /// Native window handle.
        /// </summary>
        private IntPtr _handle;

        /// <summary>
        /// Ping instance.
        /// </summary>
        private readonly Ping _ping = new Ping();

        /// <summary>
        /// Ping value list.
        /// </summary>
        private FixedSizeQueue<int> _data;

        /// <summary>
        /// Timer to invoke ping.
        /// </summary>
        private Timer _timer;

        /// <summary>
        /// Timer interval.
        /// </summary>
        private int _interval = 1000;

        /// <summary>
        /// UI fill color.
        /// </summary>
        private SolidColorBrush _fill;

        /// <summary>
        /// UI rotation animation storyboard.
        /// </summary>
        private Storyboard _rotator;

        /// <summary>
        /// UI change color animation.
        /// </summary>
        private ColorAnimation _changeColor;

        /// <summary>
        /// JPing setting
        /// </summary>
        private Setting _setting;

        #endregion


        #region Initialize and Finalize

        /// <summary>
        /// Initialize main UI.
        /// </summary>
        public MainWindow()
        {
            InitializeComponent();
            LoadSetting();
            InitializeTimer();
            InitializeAnimation();
        }

        /// <summary>
        /// Initialize ping timer.
        /// </summary>
        private void InitializeTimer()
        {
            _timer = new Timer(_setting.Interval * MS_2_S) { Enabled = true };
            _timer.Elapsed += Timer_Elapsed;
            _data = new FixedSizeQueue<int>(_setting.Average);
        }

        /// <summary>
        /// Initialize animation related objects.
        /// </summary>
        private void InitializeAnimation()
        {
            _fill = FindResource("FillBrush") as SolidColorBrush;
            _changeColor = new ColorAnimation() { Duration = TimeSpan.FromMilliseconds(_setting.Interval * MS_2_S) };
            _rotator = FindResource("RotationAnimation") as Storyboard;
            _rotator.Begin(indicator, true);
        }

        /// <summary>
        /// Get window handle when UI loaded.
        /// </summary>
        private void Window_Loaded(object sender, RoutedEventArgs e)
        {
            _handle = new WindowInteropHelper(this).EnsureHandle();
        }

        /// <summary>
        /// Save setting when UI closed.
        /// </summary>
        private void Window_Closed(object sender, EventArgs e)
        {
            SaveSetting();
        }

        #endregion


        #region Load and Save Setting

        /// <summary>
        /// Load setting from registry.
        /// </summary>
        private void LoadSetting()
        {
            try
            {
                if (Registry.GetValue(REG_PATH, REG_KEY, null) is string s)
                {
                    XmlSerializer xml = new XmlSerializer(typeof(Setting));
                    _setting = xml.Deserialize(new MemoryStream(Encoding.UTF8.GetBytes(s))) as Setting;
                }
            }
            catch { }
            if (_setting == null)
                _setting = new Setting();
            _setting.PropertyChanged += Setting_PropertyChanged;
            number.Visibility = _setting.DisplayPingValue ? Visibility.Visible : Visibility.Hidden;
            Left = _setting.Left;
            Top = _setting.Top;
        }

        /// <summary>
        /// Save setting to registry.
        /// </summary>
        private void SaveSetting()
        {
            _setting.PropertyChanged -= Setting_PropertyChanged;
            _setting.Left = Left;
            _setting.Top = Top;

            try
            {
                XmlSerializer xml = new XmlSerializer(typeof(Setting));
                using (MemoryStream ms = new MemoryStream())
                {
                    xml.Serialize(ms, _setting);
                    Registry.SetValue(REG_PATH, REG_KEY, Encoding.UTF8.GetString(ms.ToArray()));
                }
            }
            catch { }
        }

        /// <summary>
        /// Handling setting changed event.
        /// </summary>
        private void Setting_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            switch (e.PropertyName)
            {
                case "Average":
                    _data.Capacity = _setting.Average;
                    break;
                case "DisplayPingValue":
                    number.Visibility = _setting.DisplayPingValue ? Visibility.Visible : Visibility.Hidden;
                    break;
                case "Interval":
                    _interval = (int)(_setting.Interval * MS_2_S);
                    _timer.Interval = _interval;
                    _changeColor.Duration = TimeSpan.FromMilliseconds(_interval);
                    _rotator.SetSpeedRatio(indicator, 1 / Math.Sqrt(_setting.Interval));
                    break;
                case "Target":
                    _data.Clear();
                    break;
                case "TextColor":
                    number.Foreground = new SolidColorBrush(_setting.TextColor);
                    break;
                case "Topmost":
                    Topmost = _setting.Topmost;
                    break;
            }
        }

        #endregion


        #region Ping related Methods

        /// <summary>
        /// Invoke ping repeatly.
        /// </summary>
        private void Timer_Elapsed(object sender, ElapsedEventArgs e)
        {
            int ping = Ping();
            _data.Enqueue(ping >= 0 ? ping : _interval + 1);
            double p = _data.Average();
            string s;
            Color c;
            if (p >= _interval)
            {
                s = "X";
                c = _setting.UseTimeoutColor ? _setting.TimeoutColor : _setting.UpperColor;
            }
            else
            {
                s = p.ToString("F0");
                if (p >= _setting.UpperGate)
                    c = _setting.UpperColor;
                else if (p <= _setting.LowerGate)
                    c = _setting.LowerColor;
                else
                    c = _setting.GetColor(p);
            }
            Dispatcher.Invoke(() =>
            {
                number.Content = s;
                ChangeColor(c);
            });
        }

        /// <summary>
        /// Ping the target host and get the result time in milliseconds (ping value).
        /// </summary>
        /// <returns>Ping value</returns>
        private int Ping()
        {
            try
            {
                PingReply res = _ping.Send(_setting.Target, _interval);
                if (res.Status == IPStatus.Success)
                    return (int)res.RoundtripTime;
                else
                    return -1;
            }
            catch (PingException) { return -1; }
            catch (InvalidOperationException) { return -1; }
        }

        /// <summary>
        /// Change the UI color with animation.
        /// </summary>
        /// <param name="color">New color</param>
        private void ChangeColor(Color color)
        {
            _changeColor.To = color;
            _fill.BeginAnimation(SolidColorBrush.ColorProperty, _changeColor);
        }

        #endregion


        #region UI Event Handlers

        /// <summary>
        /// Move UI.
        /// </summary>
        private void Window_MouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            NativeMethods.ReleaseCapture();
            NativeMethods.SendMessage(_handle, NativeMethods.WindowsMessage.NCLButtonDown, (IntPtr)2, IntPtr.Zero);
        }

        /// <summary>
        /// Close UI.
        /// </summary>
        private void Window_MouseDoubleClick(object sender, MouseButtonEventArgs e)
        {
            Close();
        }

        /// <summary>
        /// Close UI.
        /// </summary>
        private void Exit_Click(object sender, RoutedEventArgs e)
        {
            Close();
        }

        /// <summary>
        /// Open setting window.
        /// </summary>
        private void Setting_Click(object sender, RoutedEventArgs e)
        {
            int i = (int)(sender as System.Windows.Controls.Control).Tag;
            SettingWindow w = new SettingWindow(_setting, i) { Owner = this };
            w.ShowDialog();
        }

        #endregion
    }
}
